/* Output from p2c, the Pascal-to-C translator */
/* From input file "M2/KdTree.mi" */


/*
 *  Find nearest neighbors to a given input point quickly.
 */


#include <p2c/p2c.h>


#ifndef DynMatrix_H
#include "DynMatrix.h"
#endif

#ifndef DynArray_H
#include "DynArray.h"
#endif

#ifndef KdTree_H
#include "KdTree.h"
#endif

/* These routines are NOT re-entrant!!! */

#ifndef Util_H
#include "Util.h"
#endif


#define MaxBucketSize   15
/* Maximum number of nodes in a bucket */
#define MaxDim          30


/* Saved param's of input tree, to avoid passing up and down */
Static float **globmat;
/*  globindex       :       BucketPtr; */
Static unsigned long globlength, globwidth;

Static unsigned long globneigh;   /* how many neigbors to search for */

Static unsigned long *bestindexes;
Static float *bestdistances;
/* Dynamically allocated variables---best points found so far */

Static float globcenter[MaxDim + 1];

/* when searching by distance these are meaningful as well */
Static unsigned long globnfound;
/* when searching by dis, how many
                                       we've found so far */
Static float globmaxdissqr;   /* how far out to search squared*/

Static unsigned long globcenter_index, globdecorrel_time;

Static boolean globworry_about_decorrelation;


/* Local */
Static void KdTree_SelectSplit(unsigned long *inputindices,
  unsigned long numpoints, unsigned long *whichdim, float *where)
{
    /* Choose a dimension to split the bucket on, and say where it is */
    unsigned long maxindex;
    float maxspread, bestmean, sum, mean, dev, tmp;

    unsigned long i, j, FORLIM;

    /* Find the coordinate with the greatest "spread" */
    maxspread = 0.0;
    bestmean = 0.0;
    maxindex = -1;

    /* ADD IN HACK to save time!   only calculate averages on 1000 points*/
    if (numpoints > 1000)
	numpoints = 1000;

    FORLIM = globwidth;
    for (i = 0; i < FORLIM; i++) {
	sum = 0.0;
	for (j = 0; j < numpoints; j++)
	    sum += globmat[inputindices[j]][i];

	mean = sum / numpoints;   /* now have the mean */

	sum = 0.0;
	for (j = 0; j < numpoints; j++) {
	    tmp = globmat[inputindices[j]][i] - mean;
	    if (tmp < 0)
		tmp = -tmp;
	    sum += tmp;
	}

	dev = sum / numpoints;
	/* mean deviation for this coordinate */

	if (dev > maxspread) {
	    maxspread = dev;
	    maxindex = i;
	    bestmean = mean;
	}
    }

    if (maxindex == -1)
	Util_FatalError("***BAD NEWS maxindex never got set!! ***");

    *whichdim = maxindex;
    *where = bestmean;

}


/* Local */
Static tNode *KdTree_CreateNode(unsigned long *inputindices,
				unsigned long numpoints)
{
    /* Create a subnode in the K-D tree, with the given input indices
       Valid points run from inputindices^[0] to [numpoints-1] */
    tNode *thisnode;
    unsigned long whichdim;
    float where;
    unsigned long i, nleft, nright;
    float val;

    thisnode = Malloc(sizeof(tNode));

    if (numpoints <= MaxBucketSize) {
	/* terminal node! */
	thisnode->nbucket = numpoints;
	thisnode->bucket = inputindices;

	thisnode->left = NULL;
	thisnode->left = NULL;
	thisnode->discriminator = 0;
	thisnode->partition = 0.0;

	/* writeln('Building bucket node with ',numpoints); */

	return thisnode;
    }  /* we have to build the node */

    where = 666.0;
    whichdim = 666;

    KdTree_SelectSplit(inputindices, numpoints, &whichdim, &where);
    /* writeln('Splitting on dimension ',whichdim,' at ',where); */
    /* choose a splitting */
    thisnode->nbucket = 0;
    thisnode->bucket = NULL;
    thisnode->discriminator = whichdim;
    thisnode->partition = where;

    /* Now, do the split */
    nleft = 0;   /* number known to be in the left side */
    nright = 0;   /*      "       "            right "   */

    /* Put all vectors less than the splitting "to the left"
       of the others---and point the left bucket to be starting
       at the beginning, and the right bucket to start at the
       splitting point */

    do {
	i = inputindices[nleft];
	val = globmat[i][whichdim];
	/* the test value */

	if (val <= where) {
	    /* it's on the left side alright */
	    nleft++;
	}  /* swap it with a right sider */
	else {
	    inputindices[nleft] = inputindices[numpoints - nright - 1];
	    inputindices[numpoints - nright - 1] = i;
	    nright++;
	}

    } while (nleft + nright != numpoints);
    if (nleft == 0) {
	printf("*** BAD NEWS nleft = 0, nright = %12lu\n", nright);
	printf("*** numpoints = %12lu whichdim = %12lu\n",
	       numpoints, whichdim);
    }

    if (nright == 0) {
	printf("*** BAD NEWS nright = 0, left = %12lu\n", nright);
	printf("*** numpoints = %12lu whichdim = %12lu\n",
	       numpoints, whichdim);
    }


    thisnode->left = KdTree_CreateNode(inputindices, nleft);
    thisnode->right = KdTree_CreateNode(&inputindices[numpoints - nright],
					nright);
    return thisnode;
}


tTopRec *KdTree_CreateKdTree(float **inputmatrix, unsigned long length,
			     unsigned long width)
{

    /* inputmatrix  :       Input matrix---Make sure you DO NOT change
                            the contents of RealMatrixPtr "behind the back"
                            off the tree routines !!!
       length       :       0 to length-1 gives valid entries
       width        :       0 to width-1 gives valid entries                 */
    tTopRec *tree;
    unsigned long i;
    unsigned long *intarr;

    tree = Malloc(sizeof(tTopRec));   /* Gotta make a structure */

    globmat = inputmatrix;
    globlength = length;
    globwidth = width;


    /* Now create a big array of integers, to be used by the buckets */

    intarr = DynArray_MakeCardArray(length);

    for (i = 0; i < length; i++)
	intarr[i] = i;
    /*  globindex := intarr; */

    tree->bigbucket = intarr;
    tree->numvectors = length;
    tree->widthvectors = width;
    tree->matrix = inputmatrix;

    tree->treehead = KdTree_CreateNode(intarr, length);
    /* this routine will recursively create everything */

    return tree;

}


/*local*/
Static void KdTree_FreeNode(tNode *node)
{
    if (node->bucket == NULL) {   /*non terminal node */
	if (node->left != NULL)
	    KdTree_FreeNode(node->left);

	if (node->right != NULL)
	    KdTree_FreeNode(node->right);
    }

    Free(node);

}


void KdTree_DestroyKdTree(tTopRec **treeptr)
{
    KdTree_FreeNode((*treeptr)->treehead);

    DynArray_FreeArray((void **)(&(*treeptr)->bigbucket));

    Free(*treeptr);
    *treeptr = NULL;

}


/*
 **** Search Routines  ******
 */


/* local */
Static void KdTree_ProcessTerminalNode(tNode *node)
{
    long i, j;
    unsigned long index_;
    float cutoff, dis, tmp;
    long FORLIM;


    FORLIM = node->nbucket;
    for (i = 0; i < FORLIM; i++) {
	/* loop over all elements of the bucket */
	cutoff = bestdistances[globneigh - 1];
	/* the farthest neighbor found so far */

	/* compute distance from search vector */
	dis = 0.0;
	index_ = node->bucket[i];
	for (j = globwidth - 1; j >= 0; j--) {
	    tmp = globcenter[j] - globmat[index_][j];
	    /* for L-INFINITY */
	    /*
	                IF tmp < 0 THEN
	                    tmp := - tmp;
	                END;

	                IF tmp > dis THEN
	                    dis := tmp;
	                END;
	     */
	    /* for EUCLIDEAN */
	    dis += tmp * tmp;


	    if (dis > cutoff)
		break;
	}  /* FOR j */


	if (dis < cutoff) {   /* have we found a new node */
	    /* Insert the element in the proper place */
	    /* try to be efficient about it as this routine uses the majority of the time
	       in the search proc. */

	    for (j = globneigh - 1; j >= 1; j--) {
		if (bestdistances[j - 1] < dis)
		    break;

		bestdistances[j] = bestdistances[j - 1];
		bestindexes[j] = bestindexes[j - 1];

	    }


	    bestdistances[j] = dis;
	    bestindexes[j] = index_;

	}


    }  /* FOR i */

}


Static void KdTree_ProcessTerminalNodeDecorrelate(tNode *node)
{
    long i, j;
    unsigned long index_;
    float cutoff, dis, tmp;
    long FORLIM;


    FORLIM = node->nbucket;
    for (i = 0; i < FORLIM; i++) {
	/* loop over all elements of the bucket */
	cutoff = bestdistances[globneigh - 1];
	/* the farthest neighbor found so far */

	/* compute distance from search vector */
	dis = 0.0;
	index_ = node->bucket[i];
	if (labs(index_ - globcenter_index) > globdecorrel_time) {
	    for (j = globwidth - 1; j >= 0; j--) {
		tmp = globcenter[j] - globmat[index_][j];
		/* for L-INFINITY */
		/*
		            IF tmp < 0 THEN
		                tmp := - tmp;
		            END;

		            IF tmp > dis THEN
		                dis := tmp;
		            END;
		 */
		/* for EUCLIDEAN */
		dis += tmp * tmp;


		if (dis > cutoff)
		    break;
	    }  /* FOR j */


	    if (dis < cutoff) {   /* have we found a new node */
		/* Insert the element in the proper place */
		/* try to be efficient about it as this routine uses the majority of the time
		   in the search proc. */

		for (j = globneigh - 1; j >= 1; j--) {
		    if (bestdistances[j - 1] < dis)
			break;

		    bestdistances[j] = bestdistances[j - 1];
		    bestindexes[j] = bestindexes[j - 1];

		}


		bestdistances[j] = dis;
		bestindexes[j] = index_;

	    }

	}  /* correlation is OK */


    }  /* FOR i */

}




/* local */
Static void KdTree_SearchNode(tNode *node)
{

    /* This is the guts of the K-D tree searching algorithm--- We use the
       new technique. */
    /* the discriminator for this node */
    float biggestfound;
    /* Best found so far */

    float partdiff, partdis;

    /* difference/distance of our center to the partition of this node */


    if (node->nbucket != 0) {
	if (globworry_about_decorrelation == false)
	    KdTree_ProcessTerminalNode(node);
	else
	    KdTree_ProcessTerminalNodeDecorrelate(node);
	return;
    }

    partdiff = globcenter[node->discriminator] - node->partition;

    /* Now we  either have to chose the L-infinity norm, or the Euclidean norm.
       We'll do L-infinity for now. */

    /* for L-INFINITY
            IF (partdiff < 0 ) THEN
                partdis := -partdiff;
            ELSE
                partdis := partdiff;
            END; */

    /* for EUCLIDEAN */
    partdis = partdiff * partdiff;


    if (partdiff < 0.0) {
	KdTree_SearchNode(node->left);
	biggestfound = bestdistances[globneigh - 1];

	if (partdis < biggestfound)
	    KdTree_SearchNode(node->right);

	return;
    }

    KdTree_SearchNode(node->right);
    biggestfound = bestdistances[globneigh - 1];

    if (partdis < biggestfound)
	KdTree_SearchNode(node->left);



    /* if a bucket? */
}


void KdTree_SearchKdTree(tTopRec *treeptr, float *point, unsigned long nneigh,
			 unsigned long *indexes, float *distances)
{

    /* point        :       The center to search around .
       nneight      :       How many neighbors.  Note: if point is in
                            the dataset, then the 1st returned will
                            be the point itself.
       indexes      :       returned indexes of the neighbors (0..length-1)
       distances    :        "        distances

                            These last 2 must ALREADY be allocated.
       The top level searching routine.
     */
    unsigned long i, FORLIM;


    globworry_about_decorrelation = false;

    /* set up global arrays */
    bestindexes = indexes;
    bestdistances = distances;
    globneigh = nneigh;

    globlength = treeptr->numvectors;
    globwidth = treeptr->widthvectors;
    globmat = treeptr->matrix;



    FORLIM = treeptr->widthvectors;
    for (i = 0; i < FORLIM; i++)
	globcenter[i] = point[i];

    for (i = 0; i <= nneigh; i++) {
	bestdistances[i] = 1e10;
	bestindexes[i] = -666;
    }

    KdTree_SearchNode(treeptr->treehead);
    /* As the tree is searched, the bestindexes and bestdistances arrays will be updated */


}


void KdTree_SearchKdTreeCorrel(tTopRec *treeptr, float *point,
  unsigned long nneigh, unsigned long center_index,
  unsigned long decorrel_time, unsigned long *indexes, float *distances)
{

    /* point        :       The center to search around .
       nneigh      :       How many neighbors.  Note: if point is in
                            the dataset, then the 1st returned will
                            be the point itself.
       center_index :
       decorrel_time:       Only return points whose |index - center_index|  < decorreltime
       indexes      :       returned indexes of the neighbors (0..length-1)
       distances    :        "        distances

                            These last 2 must ALREADY be allocated.
       The top level searching routine.
     */
    unsigned long i, FORLIM;


    globworry_about_decorrelation = true;
    globcenter_index = center_index;
    /* set up global arrays */
    bestindexes = indexes;
    bestdistances = distances;
    globneigh = nneigh;

    globlength = treeptr->numvectors;
    globwidth = treeptr->widthvectors;
    globmat = treeptr->matrix;



    FORLIM = treeptr->widthvectors;
    for (i = 0; i < FORLIM; i++)
	globcenter[i] = point[i];

    for (i = 0; i <= nneigh; i++) {
	bestdistances[i] = 1e10;
	bestindexes[i] = -666;
    }

    KdTree_SearchNode(treeptr->treehead);
    /* As the tree is searched, the bestindexes and bestdistances arrays will be updated */


}




/*
 *  Search routines for maximum distance not fixed # neighbors
 */

/* local */
Static void KdTree_ProcessTerminalNodeDis(tNode *node)
{
    long i, j;
    unsigned long index_;
    float dis, tmp;
    long FORLIM;


    FORLIM = node->nbucket;
    for (i = 0; i < FORLIM; i++) {
	/* compute distance from search vector */
	dis = 0.0;
	index_ = node->bucket[i];
	for (j = globwidth - 1; j >= 0; j--) {
	    tmp = globcenter[j] - globmat[index_][j];
	    /* for L-INFINITY
	                IF tmp < 0 THEN
	                    tmp := - tmp;
	                END;

	                IF tmp > dis THEN
	                    dis := tmp;
	                END;
	     */

	    /* for EUCLIDEAN */
	    dis += tmp * tmp;


	    if (dis > globmaxdissqr)
		break;
	}  /* FOR j */


	if (dis < globmaxdissqr) {   /* have we found a new node */
	    /* Insert the element in the proper place */
	    /* try to be efficient about it as this routine uses the majority
	       of the time in the search proc. */

	    globnfound++;   /* bump it up */
	    /* 0 to this number - 1 */

	    for (j = globnfound - 1; j >= 1; j--) {
		if (bestdistances[j - 1] < dis)
		    break;

		bestdistances[j] = bestdistances[j - 1];
		bestindexes[j] = bestindexes[j - 1];


	    }


	    bestdistances[j] = dis;
	    bestindexes[j] = index_;

	}


    }  /* FOR i */
    /* loop over all elements of the bucket */


}


/* local */
Static void KdTree_SearchNodeDis(tNode *node)
{

    /* This is the guts of the K-D tree searching algorithm--- We use the
       new technique. */
    /* the discriminator for this node */
    float biggestfound;
    /* Best found so far */

    float partdiff, partdis;

    /* difference/distance of our center to the partition of this node */


    if (node->nbucket != 0) {
	KdTree_ProcessTerminalNodeDis(node);
	return;
    }

    partdiff = globcenter[node->discriminator] - node->partition;

    /* Now we  either have to chose the L-infinity norm, or the Euclidean norm.
       We'll do L-infinity for now. */

    /* for L-INFINITY
            IF (partdiff < 0 ) THEN
                partdis := -partdiff;
            ELSE
                partdis := partdiff;
            END;
     */

    /* for EUCLIDEAN */
    partdis = partdiff * partdiff;

    if (partdiff < 0.0) {
	KdTree_SearchNodeDis(node->left);
	biggestfound = bestdistances[globneigh - 1];

	if (partdis < globmaxdissqr)
	    KdTree_SearchNodeDis(node->right);

	return;
    }

    KdTree_SearchNodeDis(node->right);
    biggestfound = bestdistances[globneigh - 1];

    if (partdis < globmaxdissqr)
	KdTree_SearchNodeDis(node->left);



    /* if a bucket? */
}



void KdTree_SearchKdTreeDis(tTopRec *treeptr, float *point, float maxdis,
			    unsigned long *nneigh, unsigned long *indexes,
			    float *distances)
{
    /* returned  only */
    unsigned long i, FORLIM;


    globworry_about_decorrelation = false;

    /* set up global arrays */
    bestindexes = indexes;
    bestdistances = distances;

    globnfound = 0;
    globmaxdissqr = maxdis;

    globlength = treeptr->numvectors;
    globwidth = treeptr->widthvectors;
    globmat = treeptr->matrix;

    FORLIM = treeptr->widthvectors;
    for (i = 0; i < FORLIM; i++)
	globcenter[i] = point[i];

    KdTree_SearchNodeDis(treeptr->treehead);
    /* As the tree is searched, the bestindexes and bestdistances arrays will be updated */

    *nneigh = globnfound;

}




/* End. */
