/* Header for module KdTree, generated by p2c */
#ifndef KdTree_H
#define KdTree_H


/*
 *  Find nearest neighbors to a given input point quickly.
 */

#ifndef DynMatrix_H
#include "DynMatrix.h"
#endif

#ifndef DynArray_H
#include "DynArray.h"
#endif


#ifdef KdTree_G
# define vextern
#else
# define vextern extern
#endif


typedef unsigned long Bucket[100001L];

typedef struct tNode {
    unsigned long discriminator;
    /* which dimension to judge on? */
    float partition;
    /* what's the cutoff value */

    unsigned long *bucket;
    /* indices of the vectors in this bucket */
    /* null if nonterminal node. */

    unsigned long nbucket;
    /* Number of elements in the bucket */

    struct tNode *left, *right;
    /* child nodes */
} tNode;



typedef struct tTopRec {
    long numvectors;
    /* vectors go from 0 to numvectors-1 */
    long widthvectors;
    /* the dimension of the vectors */

    float **matrix;
    /* a dynamically allocated matrix */
    /* that holds the vectors */

    tNode *treehead;
    /* top of the tree */

    unsigned long *bigbucket;
    /* A big Dynamic array--individual buckets */
    /* will point into here. */
} tTopRec;



extern tTopRec *KdTree_CreateKdTree(float **inputmatrix, unsigned long length,
				    unsigned long width);

/* inputmatrix  :       Input matrix---Make sure you DO NOT change
                        the contents of RealMatrixPtr "behind the back"
                        off the tree routines !!!
   length       :       0 to length-1 gives valid entries
   width        :       0 to width-1 gives valid entries                 */


extern void KdTree_DestroyKdTree(tTopRec **treeptr);
/* Nuke it. */

extern void KdTree_SearchKdTree(tTopRec *treeptr, float *point,
  unsigned long nneigh, unsigned long *indexes, float *distances);

/* point        :       The center to search around .
   nneight      :       How many neighbors.  Note: if point is in
                        the dataset, then the 1st returned will
                         be the point itself.
   indexes      :       returned indexes of the neighbors (0..length-1)
   distances    :        "        distances

        These last 2 must ALREADY be allocated.
 */

extern void KdTree_SearchKdTreeCorrel(tTopRec *treeptr, float *point,
  unsigned long nneigh, unsigned long center_index,
  unsigned long decorrel_time, unsigned long *indexes, float *distances);

/*  Same as above, but only returns those points whose absolute difference in
    index from the center point are strictly
    larger than decorrel_time.  If decorrel_time is < 0 then should have
    the same effect as SearchKdTree.
 */


extern void KdTree_SearchKdTreeDis(tTopRec *treeptr, float *point,
  float maxdis, unsigned long *nneigh, unsigned long *indexes,
  float *distances);

/* point        :       The center to search around.
   maxdis:The Maximum distance to search out to.
   nneigh       :       How many neighbors we found.  Note: if point is in
                        the dataset, then the 1st returned will
                         be the point itself.
   indexes      :       returned indexes of the neighbors (0..length-1)
   distances    :        "        distances

        These last 2 must ALREADY be allocated.
 */






#undef vextern

#endif /*KdTree_H*/

/* End. */
