/* Output from p2c, the Pascal-to-C translator */
/* From input file "M2/Math.mi" */


#include <p2c/p2c.h>


#ifndef DynArray_H
#include "DynArray.h"
#endif



void Math_HeapSortReal(unsigned long N, float *A)
{
    /* Sort A[0..N-1] into ascending numerical order */
    /* From numerical recipes */
    unsigned long l, j, ir, i;
    float rra;


    l = N / 2 + 1;
    ir = N;

    do {
	if (l > 1) {
	    l--;
	    rra = A[l - 1];
	} else {
	    rra = A[ir - 1];
	    A[ir - 1] = A[0];
	    ir--;
	    if (ir == 1) {
		A[0] = rra;
		return;
	    }
	}  /* if */


	i = l;
	j = l * 2;
	while (j <= ir) {
	    if (j < ir && A[j - 1] < A[j])
		j++;

	    if (rra < A[j - 1]) {
		A[i - 1] = A[j - 1];
		i = j;
		j += j;
	    } else
		j = ir + 1;
	}  /* while */
	A[i - 1] = rra;
    } while (true);   /* i.e. always loop.  Will exit with return */

}


void Math_HeapSortReal2(unsigned long N, float *A, float *B)
{
    /* Sort A[0..N-1] into ascending numerical order,
       while B[0..N-1] goes along for the ride */
    /* From numerical recipes */
    unsigned long l, j, ir, i;
    float rra, rrb;


    l = N / 2 + 1;
    ir = N;

    do {
	if (l > 1) {
	    l--;
	    rra = A[l - 1];
	    rrb = B[l - 1];
	} else {
	    rra = A[ir - 1];
	    rrb = B[ir - 1];
	    A[ir - 1] = A[0];
	    B[ir - 1] = B[0];
	    ir--;
	    if (ir == 1) {
		A[0] = rra;
		B[0] = rrb;
		return;
	    }
	}  /* if */


	i = l;
	j = l * 2;
	while (j <= ir) {
	    if (j < ir && A[j - 1] < A[j])
		j++;

	    if (rra < A[j - 1]) {
		A[i - 1] = A[j - 1];
		B[i - 1] = B[j - 1];
		i = j;
		j += j;
	    } else
		j = ir + 1;
	}  /* while */
	A[i - 1] = rra;
	B[i - 1] = rrb;
    } while (true);   /* i.e. always loop.  Will exit with return */

}


void Math_InsertSortCardDouble(unsigned long N, unsigned long *A, double *B)
{

    /* Sort from 0 TO N-1 */
    unsigned long i, j, a;
    double b;


    for (j = 1; j < N; j++) {
	a = A[j];
	b = B[j];

	i = j;

	while (i > 0 && A[i - 1] > a) {
	    A[i] = A[i - 1];
	    B[i] = B[i - 1];
	    i--;
	}

	A[i] = a;
	B[i] = b;


    }  /* FOR j */

}




/* End. */
