DEFINITION MODULE KdTree;
(*
 *  Find nearest neighbors to a given input point quickly.
 *)

FROM    DynMatrix       IMPORT
    RealMatrixPtr;
FROM    DynArray        IMPORT
    CardArrayPtr,RealArrayPtr;

TYPE
    Bucket      =       ARRAY[0..100000] OF CARDINAL;
    BucketPtr   =     POINTER TO Bucket;

    tNodePtr    =       POINTER TO tNode;
    tNode       = RECORD
	discriminator   :       CARDINAL;
			      (* which dimension to judge on? *)
	partition       :       REAL;
			      (* what's the cutoff value *)

	bucket          :       BucketPtr;
			      (* indices of the vectors in this bucket *)
			      (* null if nonterminal node. *)

	nbucket         :       CARDINAL;
			      (* Number of elements in the bucket *)

	left            :       tNodePtr;
	right           :       tNodePtr;
			      (* child nodes *)
    END;



    tTopRec     = RECORD
	numvectors      :       INTEGER;
			      (* vectors go from 0 to numvectors-1 *)
	widthvectors    :       INTEGER;
			      (* the dimension of the vectors *)

	matrix          :       RealMatrixPtr;
			      (* a dynamically allocated matrix *)
			      (* that holds the vectors *)

	treehead        :       tNodePtr;
			      (* top of the tree *)

	bigbucket       :       BucketPtr;
			      (* A big Dynamic array--individual buckets *)
			      (* will point into here. *)
    END;

    KdTreePtr   = POINTER TO tTopRec;


PROCEDURE CreateKdTree(inputmatrix     :       RealMatrixPtr;
		       length          :       CARDINAL;
		       width           :       CARDINAL) : KdTreePtr;

(* inputmatrix  :       Input matrix---Make sure you DO NOT change
			the contents of RealMatrixPtr "behind the back"
			off the tree routines !!!
   length       :       0 to length-1 gives valid entries
   width        :       0 to width-1 gives valid entries                 *)


PROCEDURE DestroyKdTree(VAR treeptr    : KdTreePtr );
(* Nuke it. *)

PROCEDURE SearchKdTree(treeptr         : KdTreePtr;
		       point           : RealArrayPtr;
		       nneigh          : CARDINAL;
		       indexes         : CardArrayPtr;
		       distances       : RealArrayPtr );

(* point        :       The center to search around .
   nneight      :       How many neighbors.  Note: if point is in
			the dataset, then the 1st returned will
			 be the point itself.
   indexes      :       returned indexes of the neighbors (0..length-1)
   distances    :        "        distances

	These last 2 must ALREADY be allocated.
 *)

PROCEDURE SearchKdTreeCorrel(treeptr         : KdTreePtr;
			     point           : RealArrayPtr;
			     nneigh          : CARDINAL;
			     center_index    : CARDINAL;
			     decorrel_time   : CARDINAL;
			     indexes         : CardArrayPtr;
			     distances       : RealArrayPtr );

(*  Same as above, but only returns those points whose absolute difference in
    index from the center point are strictly 
    larger than decorrel_time.  If decorrel_time is < 0 then should have
    the same effect as SearchKdTree.
 *)


PROCEDURE SearchKdTreeDis(treeptr	: KdTreePtr;
		          point		: RealArrayPtr;
			  maxdis	: REAL;
		          VAR nneigh	: CARDINAL;
		          indexes	: CardArrayPtr;
		           distances	: RealArrayPtr );

(* point        :       The center to search around.
   maxdis	:	The Maximum distance to search out to.
   nneigh       :       How many neighbors we found.  Note: if point is in
			the dataset, then the 1st returned will
			 be the point itself.
   indexes      :       returned indexes of the neighbors (0..length-1)
   distances    :        "        distances

	These last 2 must ALREADY be allocated.
 *)






END KdTree.
